﻿Module kml_circle
    '# Copyright (c) 2007 Nick Galbreath
    '#
    '# Permission is hereby granted, free of charge, to any person obtaining a copy
    '# of this software and associated documentation files (the "Software"), to deal
    '# in the Software without restriction, including without limitation the rights
    '# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    '# copies of the Software, and to permit persons to whom the Software is
    '# furnished to do so, subject to the following conditions:
    '#
    '# The above copyright notice and this permission notice shall be included in
    '# all copies or substantial portions of the Software.
    '#
    '# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    '# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    '# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    '# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    '# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    '# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    '# THE SOFTWARE.
    '#
    '
    '#
    '# Version 2 - 12-Sept-2007 Simplified XML output
    '#                          Added commandline interface
    '# Version 1 - 10-Sept-2007 Initial release
    '#
    '
    'from math import *
    Public Const CRLF As String = Chr(13) & Chr(10)
    Structure coord
        Dim latitude, longitude As Double
        Public Sub New(ByVal longitude As Double, ByVal latitude As Double)
            Me.latitude = latitude
            Me.longitude = longitude
        End Sub
        Public Sub New(ByVal x As Double, ByVal y As Double, ByVal z As Double)
            ' Convert (x,y,z) on unit sphere back to (long, lat)
            '
            ' p is vector of three elements

            If (x = 0.0) Then
                Me.longitude = Math.PI / 2.0
            Else
                Me.longitude = Math.Atan(y / x)
            End If
            Dim colatitude As Double = Math.Acos(z)
            Me.latitude = (Math.PI / 2.0 - colatitude)

            ' select correct branch of arctan
            If x < 0.0 Then
                If y <= 0.0 Then
                    Me.longitude -= Math.PI
                Else
                    Me.longitude += Math.PI
                End If
            End If

            Const DEG As Double = 180.0 / Math.PI
            Me.longitude *= DEG
            Me.latitude *= DEG
        End Sub
        Public Sub New(ByVal p As xyz)
            Me.New(p.x, p.y, p.z)
        End Sub
    End Structure

    Structure xyz
        Dim x, y, z As Double
        Public Sub New(ByVal x As Double, ByVal y As Double, ByVal z As Double)
            Me.x = x
            Me.y = y
            Me.z = z
        End Sub
        Public Sub New(ByVal longitude As Double, ByVal latitude As Double)
            Dim theta As Double = longitude
            ' spherical coordinate use "co-latitude", not "latitude"
            ' lattiude = [-90, 90] with 0 at equator
            ' co-latitude = [0, 180] with 0 at north pole
            Dim phi As Double = Math.PI / 2.0 - latitude
            Me.x = Math.Cos(theta) * Math.Sin(phi)
            Me.y = Math.Sin(theta) * Math.Sin(phi)
            Me.z = Math.Cos(phi)
        End Sub
    End Structure
    Private Function toRad(ByVal deg As Double) As Double ' convert degrees to radians
        Return deg * Math.PI / 180
    End Function
    Private Function Earth_radius(ByVal latitude As Double) As Double

        '        Const Earth_rad As Double = 6378137 ' Mean Radius of Earth, meters
        ' WGS84 Spheroid approximation... linear interpolation of radius between
        ' Polar radius and equatorial radius, depending latitude
        Const Equatorial_Radius = 6378137
        Const Polar_Radius = 6356752
        Return (Equatorial_Radius * (90 - Math.Abs(latitude)) + Polar_Radius * Math.Abs(latitude)) / 90
    End Function
    Public Function bearing(ByVal pt1 As coord, ByVal pt2 As coord) As Double

        'Initial bearing which if followed in a straight line along a great-circle arc 
        '(orthodrome) will take you from the start point to the end point.
        ' Parameters are given in lat/long degrees
        ' Returns result in radians in the range +/-PI
        Dim lat1 As Double = toRad(pt1.latitude)
        Dim lat2 As Double = toRad(pt2.latitude)
        Dim dLon As Double = toRad(pt2.longitude - pt1.longitude)
        Return Math.Atan2(Math.Cos(lat1) * Math.Sin(lat2) - Math.Sin(lat1) * Math.Cos(lat2) * Math.Cos(dLon), _
                          Math.Sin(dLon) * Math.Cos(lat2))
    End Function
    Private Function rotPoint(ByVal vec As xyz, ByVal pt As xyz, ByVal phi As Double) As coord

        ' rotate point pt, around unit vector vec by phi radians
        ' http://blog.modp.com/2007/09/rotating-point-around-vector.html

        ' remap vector for sanity
        Dim u As Double = vec.x
        Dim v As Double = vec.y
        Dim w As Double = vec.z
        Dim x As Double = pt.x
        Dim y As Double = pt.y
        Dim z As Double = pt.z

        Dim a As Double = u * x + v * y + w * z
        Dim d As Double = Math.Cos(phi)
        Dim e As Double = Math.Sin(phi)

        Return New coord((a * u + (x - a * u) * d + (v * z - w * y) * e), _
                 (a * v + (y - a * v) * d + (w * x - u * z) * e), _
                 (a * w + (z - a * w) * d + (u * y - v * x) * e))
    End Function
    Public Function arc_points(ByVal centre As coord, ByVal radius As Double, _
                             ByVal start As coord, ByVal finish As coord, _
                             Optional ByVal clockwise As Boolean = True, _
                             Optional ByVal n As Integer = 60) As coord()
        ' Calculate list of long,lat points for an arc
        ' centre: long,lat of centre of circle
        ' radius: of circle in metres
        ' start, finish: start and end points of the arc
        ' clockwise: direction of movement from start to finish
        ' n: number of sides for a full circle

        ' Returns a list of points comprising the object

        Dim start_dir As Double = bearing(centre, start)
        Dim finish_dir As Double = bearing(centre, finish)

        'Ensure that finish is within 360 degrees of start, and in the correct direction
        Dim arc_angle As Double = (bearing(centre, finish) + (Math.PI * 2) - start_dir) Mod (Math.PI * 2)
        If clockwise Then arc_angle -= Math.PI * 2

        'Calculate the number of segments for the arc, based on segments for whole circle
        n = n * Math.Abs(arc_angle) / (2 * Math.PI)

        ' Compute longitude degrees (in radians) at given latitude
        ' Use Radius of Earth at centre of arc. A good enough approximation 
        ' for arcs of small circles
        Dim r As Double = (radius / (Earth_radius(centre.latitude) * Math.Cos(toRad(centre.latitude))))

        Dim unit_vec As xyz = New xyz(toRad(centre.longitude), toRad(centre.latitude))
        Dim pts(n) As coord

        pts(0) = start

        For i As Integer = 1 To n - 1
            Dim pt_dir As Double = arc_angle * i / n
            pts(i) = rotPoint(unit_vec, New xyz(toRad(start.longitude), toRad(start.latitude)), pt_dir)
        Next i
        ' Connect to finish point exactly
        pts(n) = finish
        Return pts
    End Function
    Public Function circle_points(ByVal centre As coord, ByVal radius As Double, _
                             Optional ByVal n As Integer = 60, _
                             Optional ByVal offset As Double = 0) As coord()
        ' get raw list of points in long,lat format

        ' radius: radius of polygon in metres
        ' n: number of sides
        ' offset: rotate polygon by number of degrees

        ' Returns a list of points comprising the object

        ' Mean Radius of Earth, meters
        Dim offsetRadians As Double = toRad(offset)
        ' compute longitude degrees (in radians) at given latitude
        ' Use Radius of Earth at centre of circle. A good enough approximation for 
        ' small circles
        Dim r As Double = (radius / (Earth_radius(centre.latitude) * Math.Cos(toRad(centre.latitude))))

        Dim unit_vec As xyz = New xyz(toRad(centre.longitude), toRad(centre.latitude))
        Dim pt As xyz = New xyz(toRad(centre.longitude) + r, toRad(centre.latitude))
        Dim pts(n + 1) As coord

        For i As Integer = 0 To n
            pts(i) = rotPoint(unit_vec, pt, offsetRadians + (2.0 * Math.PI / n) * i)
        Next i
        ' connect to starting point exactly -- not sure if required, but seems to help when the polygon is not filled
        pts(n + 1) = pts(0)
        Return pts
    End Function

    Public Function kml_regular_polygon(ByVal centre As coord, ByVal meters As Double, _
                                        ByVal altitude As Double, _
                                        ByVal extrude As Integer, _
                                        Optional ByVal alt_mode As String = "absolute", _
                                        Optional ByVal segments As Integer = 30, _
                                        Optional ByVal offset As Double = 0) As String

        ' Regular polygon
        ' (longitude, latitude) in decimal degrees
        ' meters is radius in meters
        ' segments is number of sides, > 20 looks like a circle
        ' offset, rotate polygon by a number of degrees

        ' returns a string suitable for adding to a KML file.

        ' You may want to edit this function to change "extrude" and other XML nodes.

        Dim s As String = "<Polygon>" & CRLF
        s &= "  <extrude>" & CStr(extrude) & "</extrude>" & CRLF
        s &= "  <altitudeMode>" & alt_mode & "</altitudeMode>" & CRLF
        s &= "  <outerBoundaryIs><LinearRing><coordinates>" & CRLF
        For Each p As coord In circle_points(centre, meters, segments, offset)
            s &= "    " & p.longitude & "," & p.latitude & "," & altitude & CRLF
        Next p
        s &= "  </coordinates></LinearRing></outerBoundaryIs>" & CRLF
        s &= "</Polygon>" & CRLF
        Return s
    End Function
    Public Function kml_star(ByVal centre As coord, ByVal outer As Double, ByVal inner As Double, Optional ByVal segments As Integer = 10, Optional ByVal offset As Double = 0)

        ' Make a "star" or "burst" pattern

        ' (long, lat) = center point
        ' outer = radius in meters
        ' inner = radius in meters, typically < outer
        ' segments = number of "points" on the star
        ' offset = rotate by degrees

        ' Returns a XML snippet suitable for adding into a KML file

        Dim opts() As coord = circle_points(centre, outer, segments, offset)
        Dim ipts() As coord = circle_points(centre, inner, segments, offset + 180.0 / segments)

        ' interweave the outer and inner points (I'm sure there is a better way)
        Dim pts(opts.Length * 2 + 2) As coord
        For i As Integer = 0 To opts.Length - 1
            pts(i * 2) = opts(i)
            pts(i * 2 + 1) = ipts(i)
        Next i
        Dim s As String = "<Polygon>" & CRLF
        s &= "  <outerBoundaryIs><LinearRing><coordinates>" & CRLF
        For Each p As coord In pts
            s &= "    " & p.longitude & "," & p.latitude & CRLF
        Next p
        s &= "  </coordinates></LinearRing></outerBoundaryIs>" & CRLF
        s &= "</Polygon>" & CRLF
        Return s
    End Function
    '   s += "  <altitudeMode>clampToGround</altitudeMode>\n"

    '   s = kml_star(centre, radius, inner, sides, offset)
    '   s = kml_regular_polygon(centre, radius, sides, offset)

    'kml_regular_polygon((-81.0, 14.2), 10000, 4)
    '<Polygon>
    '  <outerBoundaryIs><LinearRing><coordinates>
    '   -80.9073366544,14.2
    '   -81.0000183885,14.2898320472
    '   -81.0926633311,14.1999643609
    '   -80.9999816261,14.1101679528
    '   -80.9073366544,14.2
    '</coordinates></LinearRing></outerBoundaryIs>
    '</Polygon>

End Module
